/*****************************************************************************
 * posterize.c : video posterization filter
 *****************************************************************************
 * Copyright (C) 2000-2007 the VideoLAN team
 * $Id $
 *
 * Author: Philippe Morin <phmorin@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <stdlib.h>                                      /* malloc(), free() */
#include <string.h>

#include <math.h>                                        /* rint() */

#include <vlc/vlc.h>
#include <vlc_vout.h>

#include "vlc_filter.h"

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int  Create    ( vlc_object_t * );
static void Destroy   ( vlc_object_t * );

static picture_t *Filter( filter_t *, picture_t * );

static int PosterizeCallback( vlc_object_t *p_this, char const *psz_var,
                           vlc_value_t oldval, vlc_value_t newval,
                           void *p_data );

#define LEVEL_TEXT  N_("Level")
#define LEVEL_LONGTEXT N_("Level from 2 up to 256")

#define FILTER_PREFIX "posterize-"

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
vlc_module_begin();
    set_description( _("Posterize video filter") );
    set_shortname( _( "Posterize" ));
    set_capability( "video filter2", 0 );
    set_category( CAT_VIDEO );
    set_subcategory( SUBCAT_VIDEO_VFILTER );

    add_integer_with_range( FILTER_PREFIX "level", 0, 2, 256, NULL,
        LEVEL_TEXT, LEVEL_LONGTEXT, VLC_FALSE );

    add_shortcut( "posterize" );
    set_callbacks( Create, Destroy );
vlc_module_end();

static const char *ppsz_filter_options[] = {
    "level", NULL
};

/*****************************************************************************
 * filter_sys_t
 *****************************************************************************/
struct filter_sys_t
{
    int     i_level;
    uint8_t a_table[256];
};

/*****************************************************************************
 * Create: allocates posterize video filter
 *****************************************************************************/
static int Create( vlc_object_t *p_this )
{
    filter_t *p_filter = (filter_t *)p_this;
    filter_sys_t *p_sys;
    int i;

    /* Allocate structure */
    p_filter->p_sys = malloc( sizeof( filter_sys_t ) );
    if( p_filter->p_sys == NULL )
    {
        msg_Err( p_filter, "out of memory" );
        return VLC_ENOMEM;
    }
    p_sys = p_filter->p_sys;

    config_ChainParse( p_filter, FILTER_PREFIX, ppsz_filter_options,
                       p_filter->p_cfg );

    p_sys->i_level = var_CreateGetIntegerCommand( p_filter,
                                                  FILTER_PREFIX "level" );

    var_AddCallback( p_filter, FILTER_PREFIX "level", PosterizeCallback, p_sys );

    p_filter->pf_video_filter = Filter;

    // init lookup table
    float step = 255.0 / (p_sys->i_level-1.0);
    for (i=0; i < 256; i++) {
	p_sys->a_table[i] = (int)(rint(i/step)*step);
    }
    
    return VLC_SUCCESS;
}

/*****************************************************************************
 * Destroy: destroy posterize filter
 *****************************************************************************/
static void Destroy( vlc_object_t *p_this )
{
    filter_t *p_filter = (filter_t *)p_this;

    free( p_filter->p_sys );
}

/*****************************************************************************
 *
 *****************************************************************************/
static picture_t *Filter( filter_t *p_filter, picture_t *p_pic )
{
    picture_t *p_outpic;
    filter_sys_t *p_sys = p_filter->p_sys;
    int i_level = p_sys->i_level;
    int i_plane;

    if( !p_pic ) return NULL;

    p_outpic = p_filter->pf_vout_buffer_new( p_filter );
    if( !p_outpic )
    {
        msg_Warn( p_filter, "can't get output picture" );
        if( p_pic->pf_release )
            p_pic->pf_release( p_pic );
        return NULL;
    }


    for( i_plane = 0 ; i_plane < p_pic->i_planes ; i_plane++ )
    {
        const int i_visible_lines = p_pic->p[i_plane].i_visible_lines;
        const int i_visible_pitch = p_pic->p[i_plane].i_visible_pitch;
        const int i_pitch         = p_pic->p[i_plane].i_pitch;
        const int i_hidden_pitch  = i_pitch - i_visible_pitch;

        const uint8_t *p_in = p_pic->p[i_plane].p_pixels;
        uint8_t *p_out = p_outpic->p[i_plane].p_pixels;
        uint8_t *p_outendline = p_out + i_visible_pitch;
        const uint8_t *p_outend = p_out + i_visible_lines * i_pitch;

        for( ; p_outendline < p_outend;
             p_out += i_hidden_pitch, p_outendline += i_pitch,p_in += i_hidden_pitch )
        {
            for( ; p_out < p_outendline;
       	         p_out++,p_in++)
       	    {
		*p_out = p_sys->a_table[*p_in];		
       	    }
        }
    }

    p_outpic->date = p_pic->date;
    p_outpic->b_force = p_pic->b_force;
    p_outpic->i_nb_fields = p_pic->i_nb_fields;
    p_outpic->b_progressive = p_pic->b_progressive;
    p_outpic->b_top_field_first = p_pic->b_top_field_first;

    if( p_pic->pf_release )
        p_pic->pf_release( p_pic );

    return p_outpic;
}

/*****************************************************************************
 *
 *****************************************************************************/
static int PosterizeCallback( vlc_object_t *p_this, char const *psz_var,
                           vlc_value_t oldval, vlc_value_t newval,
                           void *p_data )
{
    filter_sys_t *p_sys = (filter_sys_t *)p_data;
    int i;

    if( !strcmp( psz_var, "posterize-level" ) )
    {
        p_sys->i_level = newval.i_int;
 	// init lookup table
    	float step = 255.0 / (p_sys->i_level-1.0);
    	for (i=0; i < 256; i++) {
		p_sys->a_table[i] = (int)(rint(i/step)*step);
    	}
    }

    return VLC_SUCCESS;
}
